# Unit test predict_interval ForecasterDirectMultiVariate
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from sklearn.linear_model import LinearRegression
from sklearn.compose import ColumnTransformer
from sklearn.preprocessing import StandardScaler
from sklearn.preprocessing import OneHotEncoder
from skforecast.direct import ForecasterDirectMultiVariate

# Fixtures
from .fixtures_forecaster_direct_multivariate import series
from .fixtures_forecaster_direct_multivariate import exog
from .fixtures_forecaster_direct_multivariate import exog_predict

transformer_exog = ColumnTransformer(
                       [('scale', StandardScaler(), ['exog_1']),
                        ('onehot', OneHotEncoder(), ['exog_2'])],
                       remainder = 'passthrough',
                       verbose_feature_names_out = False
                   )


def test_check_interval_ValueError_when_method_is_not_valid_method():
    """
    Check ValueError is raised when `method` is not 'bootstrapping' or 'conformal'.
    """
    forecaster = ForecasterDirectMultiVariate(
        estimator=LinearRegression(), level='l1', steps=2, lags=3
    )
    forecaster.fit(series=series, store_in_sample_residuals=True)

    method = 'not_valid_method'
    err_msg = re.escape(
        f"Invalid `method` '{method}'. Choose 'bootstrapping' or 'conformal'."
    )
    with pytest.raises(ValueError, match = err_msg):
        forecaster.predict_interval(steps=1, method=method)


@pytest.mark.parametrize("interval", 
                         [0.90, [5, 95], (5, 95)], 
                         ids = lambda value: f'interval: {value}')
def test_predict_interval_output_when_in_sample_residuals_exog_and_transformer(interval):
    """
    Test output of predict_interval when estimator is LinearRegression,
    2 steps are predicted, using in-sample residuals, exog is included and both
    inputs are transformed.
    """
    forecaster = ForecasterDirectMultiVariate(
                     estimator          = LinearRegression(),
                     level              = 'l1',
                     steps              = 2,
                     lags               = 3,
                     transformer_series = StandardScaler(),
                     transformer_exog   = transformer_exog
                 )
    forecaster.fit(series=series, exog=exog, store_in_sample_residuals=True)
    n_boot = 250
    results = forecaster.predict_interval(
                    steps                   = 2,
                    exog                    = exog_predict,
                    method                  = 'bootstrapping',
                    interval                = interval,
                    n_boot                  = n_boot,
                    use_in_sample_residuals = True,
                    use_binned_residuals    = False
                )
    
    expected = pd.DataFrame(
                   data    = np.array([[0.61820497, 0.27569583, 0.94783844],
                                        [0.41314101, 0.06233635, 0.74277449]]),
                   columns = ['pred', 'lower_bound', 'upper_bound'],
                   index   = pd.RangeIndex(start=50, stop=52)
               )
    expected.insert(0, 'level', np.tile(['l1'], 2))
    
    pd.testing.assert_frame_equal(expected, results)


def test_predict_interval_output_when_forecaster_is_LinearRegression_steps_is_2_in_sample_residuals_False_exog_and_transformer():
    """
    Test output of predict_interval when estimator is LinearRegression,
    2 steps are predicted, using out-sample residuals, exog is included and both
    inputs are transformed.
    """
    forecaster = ForecasterDirectMultiVariate(
                     estimator          = LinearRegression(),
                     level              = 'l1',
                     steps              = 2,
                     lags               = 3,
                     transformer_series = StandardScaler(),
                     transformer_exog   = transformer_exog
                 )
    forecaster.fit(series=series, exog=exog, store_in_sample_residuals=True)
    forecaster.out_sample_residuals_ = forecaster.in_sample_residuals_
    forecaster.out_sample_residuals_by_bin_ = forecaster.in_sample_residuals_by_bin_    
    n_boot = 250
    results = forecaster.predict_interval(
                    steps                   = 2,
                    exog                    = exog_predict,
                    method                  = 'bootstrapping',
                    interval                = [5, 95],
                    n_boot                  = n_boot,
                    use_in_sample_residuals = False,
                    use_binned_residuals    = False
                )
    
    expected = pd.DataFrame(
                   data    = np.array([[0.61820497, 0.27569583, 0.94783844],
                                        [0.41314101, 0.06233635, 0.74277449]]),
                   columns = ['pred', 'lower_bound', 'upper_bound'],
                   index   = pd.RangeIndex(start=50, stop=52)
               )
    expected.insert(0, 'level', np.tile(['l1'], 2))

    pd.testing.assert_frame_equal(expected, results)


def test_predict_interval_output_when_forecaster_is_LinearRegression_steps_is_5_in_sample_residuals_is_True_binned_residuals_is_True():
    """
    Test output when estimator is LinearRegression 5 step ahead is predicted
    using in sample binned residuals.
    """
    forecaster = ForecasterDirectMultiVariate(
                     estimator          = LinearRegression(),
                     level              = 'l1',
                     steps              = 5,
                     lags               = 3,
                     transformer_series = StandardScaler()
                 )
    forecaster.fit(series=series, store_in_sample_residuals=True)
    results = forecaster.predict_interval(
        steps=5, method='bootstrapping', interval=(5, 95), 
        use_in_sample_residuals=True, use_binned_residuals=True
    )

    expected = pd.DataFrame(
                    data    = np.array(
                                [[0.58307704, 0.30689592, 0.97766641],
                                 [0.40064856, 0.04683134, 0.63486761],
                                 [0.29394488, 0.13006777, 0.50853662],
                                 [0.41007329, 0.0785547 , 0.85997779],
                                 [0.4390632 , 0.10298379, 0.83922684]]
                            ),
                    columns = ['pred', 'lower_bound', 'upper_bound'],
                    index   = pd.RangeIndex(start=50, stop=55, step=1)
                )
    expected.insert(0, 'level', np.tile(['l1'], forecaster.max_step))

    pd.testing.assert_frame_equal(results, expected)


def test_predict_interval_output_when_forecaster_is_LinearRegression_steps_is_5_in_sample_residuals_is_False_binned_residuals_is_True():
    """
    Test output when estimator is LinearRegression, steps=5, use_in_sample_residuals=False,
    binned_residuals=True.
    """
    forecaster = ForecasterDirectMultiVariate(
                     estimator          = LinearRegression(),
                     level              = 'l1',
                     steps              = 5,
                     lags               = 3,
                     transformer_series = StandardScaler()
                 )
    forecaster.fit(series=series, store_in_sample_residuals=True)
    forecaster.out_sample_residuals_ = forecaster.in_sample_residuals_
    forecaster.out_sample_residuals_by_bin_ = forecaster.in_sample_residuals_by_bin_

    results = forecaster.predict_interval(
        steps=5, method='bootstrapping', interval=(5, 95), 
        use_in_sample_residuals=False, use_binned_residuals=True
    )

    expected = pd.DataFrame(
                    data    = np.array(
                                [[0.58307704, 0.30689592, 0.97766641],
                                 [0.40064856, 0.04683134, 0.63486761],
                                 [0.29394488, 0.13006777, 0.50853662],
                                 [0.41007329, 0.0785547 , 0.85997779],
                                 [0.4390632 , 0.10298379, 0.83922684]]
                            ),
                    columns = ['pred', 'lower_bound', 'upper_bound'],
                    index   = pd.RangeIndex(start=50, stop=55, step=1)
                )
    expected.insert(0, 'level', np.tile(['l1'], forecaster.max_step))
    
    pd.testing.assert_frame_equal(results, expected)


@pytest.mark.parametrize("interval", 
                         [0.95, (2.5, 97.5)], 
                         ids = lambda value: f'interval: {value}')
def test_predict_interval_conformal_output_when_estimator_is_LinearRegression(interval):
    """
    Test predict output when using LinearRegression as estimator and StandardScaler
    and conformal prediction.
    """
    forecaster = ForecasterDirectMultiVariate(
                     estimator          = LinearRegression(),
                     level              = 'l1',
                     steps              = 3,
                     lags               = 3,
                     transformer_series = StandardScaler()
                 )
    forecaster.fit(series=series, store_in_sample_residuals=False)
    forecaster.set_in_sample_residuals(series=series)
    results = forecaster.predict_interval(
        steps=3, method='conformal', interval=interval, 
        use_in_sample_residuals=True, use_binned_residuals=False
    )

    expected = pd.DataFrame(
                   data = np.array([[ 0.63114259,  0.22697524,  1.03530995],
                                    [ 0.3800417 , -0.02412565,  0.78420905],
                                    [ 0.33255977, -0.07160758,  0.73672712]]),
                   index = pd.RangeIndex(start=50, stop=53, step=1),
                   columns = ['pred', 'lower_bound', 'upper_bound']
               )
    expected.insert(0, 'level', np.tile(['l1'], forecaster.max_step))
    
    pd.testing.assert_frame_equal(results, expected)


@pytest.mark.parametrize("interval", 
                         [0.95, (2.5, 97.5)], 
                         ids = lambda value: f'interval: {value}')
def test_predict_interval_conformal_output_when_binned_residuals(interval):
    """
    Test predict output when using LinearRegression as estimator and StandardScaler
    and conformal prediction with binned residuals.
    """
    forecaster = ForecasterDirectMultiVariate(
                     estimator          = LinearRegression(),
                     level              = 'l1',
                     steps              = 3,
                     lags               = 3,
                     transformer_series = StandardScaler()
                 )
    forecaster.fit(series=series, store_in_sample_residuals=True)
    results = forecaster.predict_interval(
        steps=3, method='conformal', interval=interval, 
        use_in_sample_residuals=True, use_binned_residuals=True
    )

    expected = pd.DataFrame(
                   data = np.array([
                              [0.63114259, 0.17603311, 1.08625208],
                              [0.3800417 , 0.12832655, 0.63175685],
                              [0.33255977, 0.08084462, 0.58427492]]),
                   index = pd.RangeIndex(start=50, stop=53, step=1),
                   columns = ['pred', 'lower_bound', 'upper_bound']
               )
    expected.insert(0, 'level', np.tile(['l1'], forecaster.max_step))
    
    pd.testing.assert_frame_equal(results, expected)


# Unit test BaseFold
# ==============================================================================
import pytest
import re
import numpy as np
import pandas as pd
from ....exceptions import IgnoredArgumentWarning
from skforecast.model_selection._split import BaseFold


valid_params = {
    "steps": 5,
    "initial_train_size": 100,
    'fold_stride': None,
    "window_size": 10,
    "differentiation": None,
    "refit": True,
    "fixed_train_size": True,
    "gap": 0,
    "skip_folds": None,
    "allow_incomplete_fold": True,
    "return_all_indexes": False,
    "verbose": True,
}


def test_basefold_validate_params_no_exception():
    """
    Test that no exceptions are raised for valid parameters.
    """
    cv = BaseFold()
    params = dict(valid_params)
    cv._validate_params(cv_name="TimeSeriesFold", **params)


def test_basefold_validate_params_raise_invalid_steps():
    """
    Test that ValueError is raised when steps is invalid.
    """
    cv = BaseFold()
    params = dict(valid_params)
    params["steps"] = 0
    msg = f"`steps` must be an integer greater than 0. Got {params['steps']}."
    with pytest.raises(ValueError, match=msg):
        cv._validate_params(cv_name="TimeSeriesFold", **params)

    params["steps"] = "invalid"
    msg = f"`steps` must be an integer greater than 0. Got {params['steps']}."
    with pytest.raises(ValueError, match=msg):
        cv._validate_params(cv_name="TimeSeriesFold", **params)

    params["steps"] = 1.0
    msg = f"`steps` must be an integer greater than 0. Got {params['steps']}."
    with pytest.raises(ValueError, match=msg):
        cv._validate_params(cv_name="TimeSeriesFold", **params)


def test_basefold_validate_params_raise_invalid_initial_train_size_TimeSeriesFold():
    """
    Test that ValueError is raised when initial_train_size is invalid 
    for TimeSeriesFold.
    """
    cv = BaseFold()
    params = dict(valid_params)
    
    # Test invalid integer
    params["initial_train_size"] = -1
    msg = (
        f"`initial_train_size` must be an integer greater than 0, "
        f"a date string, a pandas Timestamp, or None. Got {params['initial_train_size']}."
    )
    with pytest.raises(ValueError, match=msg):
        cv._validate_params(cv_name="TimeSeriesFold", **params)
    
    # Test invalid type
    params["initial_train_size"] = 1.0
    msg = (
        f"`initial_train_size` must be an integer greater than 0, a date string, "
        f"a pandas Timestamp, or None. Got {params['initial_train_size']}."
    )
    with pytest.raises(ValueError, match=msg):
        cv._validate_params(cv_name="TimeSeriesFold", **params)
    
    # Test valid None
    params["initial_train_size"] = None
    cv._validate_params(cv_name="TimeSeriesFold", **params)

    # Test valid integer
    params["initial_train_size"] = 10
    cv._validate_params(cv_name="TimeSeriesFold", **params)
    
    # Test valid date string
    params["initial_train_size"] = "2022-01-01"
    cv._validate_params(cv_name="TimeSeriesFold", **params)
    
    # Test valid pandas Timestamp
    params["initial_train_size"] = pd.Timestamp("2022-01-01")
    cv._validate_params(cv_name="TimeSeriesFold", **params)


def test_basefold_validate_params_raise_invalid_fold_stride():
    """
    Test that TypeError is raised when fold_stride is invalid.
    """
    cv = BaseFold()
    params = dict(valid_params)
    params["fold_stride"] = 0
    msg = f"`fold_stride` must be an integer greater than 0. Got {params['fold_stride']}."
    with pytest.raises(ValueError, match=msg):
        cv._validate_params(cv_name="TimeSeriesFold", **params)

    params["fold_stride"] = "invalid"
    msg = f"`fold_stride` must be an integer greater than 0. Got {params['fold_stride']}."
    with pytest.raises(ValueError, match=msg):
        cv._validate_params(cv_name="TimeSeriesFold", **params)

    params["fold_stride"] = 1.0
    msg = f"`fold_stride` must be an integer greater than 0. Got {params['fold_stride']}."
    with pytest.raises(ValueError, match=msg):
        cv._validate_params(cv_name="TimeSeriesFold", **params)


def test_basefold_validate_params_raise_invalid_refit():
    """
    Test that TypeError is raised when refit is invalid.
    """
    cv = BaseFold()
    params = dict(valid_params)
    params["refit"] = "invalid"
    msg = "`refit` must be a boolean or an integer equal or greater than 0. Got invalid."
    with pytest.raises(TypeError, match=msg):
        cv._validate_params(cv_name="TimeSeriesFold", **params)

    params["refit"] = -1
    msg = "`refit` must be a boolean or an integer equal or greater than 0. Got -1."
    with pytest.raises(TypeError, match=msg):
        cv._validate_params(cv_name="TimeSeriesFold", **params)


def test_basefold_validate_params_raise_invalid_fixed_train_size():
    """
    Test that TypeError is raised when fixed_train_size is invalid.
    """
    cv = BaseFold()
    params = dict(valid_params)
    params["fixed_train_size"] = "invalid"
    msg = "`fixed_train_size` must be a boolean: `True`, `False`. Got invalid."
    with pytest.raises(TypeError, match=msg):
        cv._validate_params(cv_name="TimeSeriesFold", **params)


def test_basefold_validate_params_raise_invalid_gap():
    """
    Test that ValueError is raised when gap is invalid.
    """
    cv = BaseFold()
    params = dict(valid_params)
    params["gap"] = -1
    msg = "`gap` must be an integer greater than or equal to 0. Got -1."
    with pytest.raises(ValueError, match=msg):
        cv._validate_params(cv_name="TimeSeriesFold", **params)

    params["gap"] = "invalid"
    msg = "`gap` must be an integer greater than or equal to 0. Got invalid."
    with pytest.raises(ValueError, match=msg):
        cv._validate_params(cv_name="TimeSeriesFold", **params)

    params["gap"] = 1.0
    msg = "`gap` must be an integer greater than or equal to 0. Got 1.0."
    with pytest.raises(ValueError, match=msg):
        cv._validate_params(cv_name="TimeSeriesFold", **params)


def test_basefold_validate_params_raise_invalid_skip_folds():
    """
    Test that ValueError is raised when skip_folds is invalid.
    """
    cv = BaseFold()
    params = dict(valid_params)
    params["skip_folds"] = 'invalid'
    msg = (
        "`skip_folds` must be an integer greater than 0, a list of "
        "integers or `None`. Got invalid."
    )
    with pytest.raises(TypeError, match=msg):
        cv._validate_params(cv_name="TimeSeriesFold", **params)

    params["skip_folds"] = -1
    msg = (
        f"`skip_folds` must be an integer greater than 0, a list of "
        f"integers or `None`. Got {params['skip_folds']}."
    )
    with pytest.raises(ValueError, match=msg):
        cv._validate_params(cv_name="TimeSeriesFold", **params)

    params["skip_folds"] = [1, 2, -1]
    msg = (
        f"`skip_folds` list must contain integers greater than or "
        f"equal to 1. The first fold is always needed to train the "
        f"forecaster. Got {params['skip_folds']}."
    )
    msg = re.escape(msg)
    with pytest.raises(ValueError, match=msg):
        cv._validate_params(cv_name="TimeSeriesFold", **params)


def test_basefold_validate_params_raise_invalid_allow_incomplete_fold():
    """
    Test that TypeError is raised when allow_incomplete_fold is invalid.
    """
    cv = BaseFold()
    params = dict(valid_params)
    params["allow_incomplete_fold"] = "invalid"
    msg = (
        f"`allow_incomplete_fold` must be a boolean: `True`, `False`. "
        f"Got {params['allow_incomplete_fold']}."
    )
    with pytest.raises(TypeError, match=msg):
        cv._validate_params(cv_name="TimeSeriesFold", **params)


def test_basefold_validate_params_raise_invalid_initial_train_size_OneStepAheadFold():
    """
    Test that ValueError is raised when initial_train_size is invalid for 
    OneStepAheadFold.
    """
    cv = BaseFold()
    params = dict(valid_params)
    
    # Test invalid integer
    params["initial_train_size"] = 0
    msg = (
        f"`initial_train_size` must be an integer greater than 0, "
        f"a date string, or a pandas Timestamp. Got {params['initial_train_size']}."
    )
    with pytest.raises(ValueError, match=msg):
        cv._validate_params(cv_name="OneStepAheadFold", **params)
    
    # Test invalid type
    params["initial_train_size"] = 1.5
    msg = (
        f"`initial_train_size` must be an integer greater than 0, a date string, or "
        f"a pandas Timestamp. Got {params['initial_train_size']}."
    )
    with pytest.raises(ValueError, match=msg):
        cv._validate_params(cv_name="OneStepAheadFold", **params)
    
    # Test valid integer
    params["initial_train_size"] = 10
    cv._validate_params(cv_name="OneStepAheadFold", **params)
    
    # Test valid date string
    params["initial_train_size"] = "2022-01-01"
    cv._validate_params(cv_name="OneStepAheadFold", **params)
    
    # Test valid pandas Timestamp
    params["initial_train_size"] = pd.Timestamp("2022-01-01")
    cv._validate_params(cv_name="OneStepAheadFold", **params)


def test_basefold_validate_params_raise_invalid_window_size():
    """
    Test that ValueError is raised when window_size is invalid.
    """
    cv = BaseFold()
    params = dict(valid_params)
    params["window_size"] = 0
    msg = f"`window_size` must be an integer greater than 0. Got {params['window_size']}."
    with pytest.raises(ValueError, match=msg):
        cv._validate_params(cv_name="TimeSeriesFold", **params)

    params["window_size"] = "invalid"
    msg = f"`window_size` must be an integer greater than 0. Got {params['window_size']}."
    with pytest.raises(ValueError, match=msg):
        cv._validate_params(cv_name="TimeSeriesFold", **params)

    params["window_size"] = 1.0
    msg = f"`window_size` must be an integer greater than 0. Got {params['window_size']}."
    with pytest.raises(ValueError, match=msg):
        cv._validate_params(cv_name="TimeSeriesFold", **params)


def test_basefold_validate_params_raise_invalid_return_all_indexes():
    """
    Test that TypeError is raised when return_all_indexes is invalid.
    """
    cv = BaseFold()
    params = dict(valid_params)
    params["return_all_indexes"] = "invalid"
    msg = (
        f"`return_all_indexes` must be a boolean: `True`, `False`. "
        f"Got {params['return_all_indexes']}."
    )
    with pytest.raises(TypeError, match=msg):
        cv._validate_params(cv_name="TimeSeriesFold", **params)


def test_basefold_validate_params_raise_invalid_differentiation():
    """
    Test that ValueError is raised when differentiation is invalid.
    """
    cv = BaseFold()
    params = dict(valid_params)
    params["differentiation"] = "invalid"
    msg = (
        f"`differentiation` must be None or an integer greater than or equal to 0. "
        f"Got {params['differentiation']}."
    )
    with pytest.raises(ValueError, match=msg):
        cv._validate_params(cv_name="TimeSeriesFold", **params)

    params["differentiation"] = -1
    msg = (
        f"`differentiation` must be None or an integer greater than or equal to 0. "
        f"Got {params['differentiation']}."
    )
    with pytest.raises(ValueError, match=msg):
        cv._validate_params(cv_name="TimeSeriesFold", **params)

    params["differentiation"] = 1.0
    msg = (
        f"`differentiation` must be None or an integer greater than or equal to 0. "
        f"Got {params['differentiation']}."
    )
    with pytest.raises(ValueError, match=msg):
        cv._validate_params(cv_name="TimeSeriesFold", **params)


def test_basefold_validate_params_raise_invalid_verbose():
    """
    Test that TypeError is raised when verbose is invalid.
    """
    cv = BaseFold()
    params = dict(valid_params)
    params["verbose"] = "invalid"
    msg = f"`verbose` must be a boolean: `True`, `False`. Got {params['verbose']}."
    with pytest.raises(TypeError, match=msg):
        cv._validate_params(cv_name="TimeSeriesFold", **params)


def test_basefold_extract_index_when_X_is_series():
    """
    Test that the index is correctly extracted when X is a pd.Series.
    """
    cv = BaseFold()
    X = pd.Series(np.arange(100))
    X.index = pd.date_range(start="2022-01-01", periods=100, freq="D")
    index = cv._extract_index(X)
    pd.testing.assert_index_equal(index, X.index)


def test_basefold_extract_index_when_X_is_pandas():
    """
    Test that the index is correctly extracted when X is a pd.DataFrame.
    """
    cv = BaseFold()
    X = pd.DataFrame(np.arange(100))
    X.index = pd.date_range(start="2022-01-01", periods=100, freq="D")
    index = cv._extract_index(X)
    pd.testing.assert_index_equal(index, X.index)


def test_basefold_extract_index_when_X_is_index():
    """
    Test that the index is correctly extracted when X is a pd.Index.
    """
    cv = BaseFold()
    X = pd.Index(pd.date_range(start="2022-01-01", periods=100, freq="D"))
    index = cv._extract_index(X)
    pd.testing.assert_index_equal(index, X)


def test_basefold_extract_index_when_X_is_dict():
    """
    Test that the index is correctly extracted when X is a dict.
    """
    cv = BaseFold()
    X = {
        "a": pd.Series(np.arange(10), index=pd.date_range(start="2022-01-01", periods=10, freq="D")),
        "b": pd.Series(np.arange(10), index=pd.date_range(start="2022-01-10", periods=10, freq="D")),
    }
    index = cv._extract_index(X)
    expected_index = pd.date_range(start="2022-01-01", end="2022-01-19", freq="D")
    pd.testing.assert_index_equal(index, expected_index)


def test_basefold_extract_index_raise_error_when_X_is_dict_with_series_with_no_valid_index():
    """
    Test that TypeError is raised when X is a dict with series with no valid index type.
    """
    cv = BaseFold()
    X = {
        "a": pd.Series(np.arange(10), index=pd.date_range(start="2022-01-01", periods=10, freq="D")),
        "b": pd.Series(np.arange(10), index=pd.Index(np.arange(10))),
    }
    msg = re.escape(
        "If `X` is a dictionary, all series must have a Pandas "
        "RangeIndex or DatetimeIndex with the same step/frequency. "
        "Review series: ['b']"
    )
    with pytest.raises(TypeError, match=msg):
        cv._extract_index(X)


def test_basefold_extract_index_raise_error_when_X_is_dict_with_series_non_with_freq():
    """
    Test that ValueError is raised when X is a dict with series with different frequencies.
    """
    cv = BaseFold()
    X = {
        "a": pd.Series(np.arange(10), index=pd.date_range(start="2022-01-01", periods=10)),
        "b": pd.Series(np.arange(10), index=pd.date_range(start="2022-01-10", periods=10)),
    }
    X["a"].index.freq = None
    X["b"].index.freq = None
    msg = re.escape(
        "If `X` is a dictionary, all series must have a Pandas "
        "RangeIndex or DatetimeIndex with the same step/frequency. "
        "Found series with no frequency or step."
    )
    with pytest.raises(ValueError, match=msg):
        cv._extract_index(X)


def test_basefold_extract_index_raise_error_when_X_is_dict_with_series_with_different_freq():
    """
    Test that ValueError is raised when X is a dict with series with different frequencies.
    """
    cv = BaseFold()
    X = {
        "a": pd.Series(np.arange(10), index=pd.date_range(start="2022-01-01", periods=10, freq="D")),
        "b": pd.Series(np.arange(10), index=pd.date_range(start="2022-01-10", periods=10, freq="2D")),
    }
    msg = re.escape(
        "If `X` is a dictionary, all series must have a Pandas "
        "RangeIndex or DatetimeIndex with the same step/frequency. "
        "Found frequencies: [<Day>, <2 * Days>]"
    )
    with pytest.raises(ValueError, match=msg):
        cv._extract_index(X)


def test_basefold_set_params_TypeError_non_dict():
    """
    Test TypeError is raised set_params method with non-dictionary input.
    """
    fold = BaseFold()
    new_params = ['not', 'a', 'dict']

    msg = (
        f"`params` must be a dictionary. Got {type(new_params)}."
    )
    with pytest.raises(TypeError, match=msg):
        fold.set_params(new_params)


def test_basefold_set_params_valid_params():
    """
    Test set_params method with valid parameters.
    """
    cv = BaseFold(initial_train_size=10, window_size=4)
    new_params = {
        'initial_train_size': 15,
        'window_size': 5,
        'verbose': False
    }
    cv.set_params(new_params)
    assert cv.initial_train_size == 15
    assert cv.window_size == 5
    assert cv.verbose is False


def test_set_params_partial_update_and_unknown_param():
    """
    Test set_params method with partial update of parameters.
    """
    cv = BaseFold(initial_train_size=10, window_size=4, verbose=True)
    new_params = {
        'initial_train_size': 15,
        'window_size': 4,
        'unknown_param': 5
    }
    warn_msg = re.escape(
        "Unknown parameters: {'unknown_param'}. They have been ignored."
    )
    with pytest.warns(IgnoredArgumentWarning, match=warn_msg):
        cv.set_params(new_params)
    
    assert cv.initial_train_size == 15
    assert cv.window_size == 4
    assert cv.verbose is True



# Unit test create_predict_X method using PyTorch backend
# ==============================================================================
import os
import re
import numpy as np
import pandas as pd
import pytest
os.environ["KERAS_BACKEND"] = "torch"
import keras
from skforecast.exceptions import DataTransformationWarning
from skforecast.deep_learning.utils import create_and_compile_model
from skforecast.deep_learning import ForecasterRnn

series = pd.DataFrame(
    {
        "1": np.arange(50),
        "2": np.arange(50),
        "3": np.arange(50),
    },
    index=pd.date_range("2020-01-01", periods=50, freq="D")
)

exog = pd.DataFrame(
    {
        "exog1": np.arange(50),
        "exog2": np.arange(50),
    },
    index=pd.date_range("2020-01-01", periods=50, freq="D")
)

exog_pred = pd.DataFrame(
    {
        "exog1": np.arange(50, 60),
        "exog2": np.arange(50, 60),
    },
    index=pd.date_range("2020-02-20", periods=10, freq="D")
)

model = create_and_compile_model(
            series=series, 
            levels=["1", "2"],    
            lags=3,           
            steps=4,              
            recurrent_layer="LSTM",
            recurrent_units=128,
            dense_units=64,
        )

model_exog = create_and_compile_model(
            series=series, 
            exog=exog,
            levels=["1", "2", "3"],    
            lags=10,           
            steps=8,              
            recurrent_layer="LSTM",
            recurrent_units=128,
            dense_units=64,
        )


def test_create_predict_X_3_steps_ahead():
    """
    Test case for create_predict_X 3 steps ahead
    """
    forecaster = ForecasterRnn(estimator=model, levels=["1", "2"], lags=3)
    forecaster.fit(series=series)

    warn_msg = re.escape(
        "The output matrix is in the transformed scale due to the "
        "inclusion of transformations in the Forecaster. "
        "As a result, any predictions generated using this matrix will also "
        "be in the transformed scale. Please refer to the documentation "
        "for more details: "
        "https://skforecast.org/latest/user_guides/training-and-prediction-matrices.html"
    )
    with pytest.warns(DataTransformationWarning, match = warn_msg):
        X_predict, exog_predict = forecaster.create_predict_X(steps=3)

    expected = (
        pd.DataFrame(
            data=np.array([
                    [0.95918367, 0.95918367, 0.95918367],
                    [0.97959184, 0.97959184, 0.97959184],
                    [1.        , 1.        , 1.        ]
                ]),
            index=['lag_3', 'lag_2', 'lag_1'],
            columns=["1", "2", "3"]
        ),
        None
    )

    pd.testing.assert_frame_equal(X_predict, expected[0])
    assert exog_predict is None


def test_create_predict_X_specific_levels():
    """
    Test case for create_predict_X with specific levels
    """
    forecaster = ForecasterRnn(estimator=model, levels=["1", "2"], lags=3)
    forecaster.fit(series=series)

    warn_msg = re.escape(
        "The output matrix is in the transformed scale due to the "
        "inclusion of transformations in the Forecaster. "
        "As a result, any predictions generated using this matrix will also "
        "be in the transformed scale. Please refer to the documentation "
        "for more details: "
        "https://skforecast.org/latest/user_guides/training-and-prediction-matrices.html"
    )
    with pytest.warns(DataTransformationWarning, match = warn_msg):
        X_predict, exog_predict = forecaster.create_predict_X(steps=None, levels=["1"])

    expected = (
        pd.DataFrame(
            data=np.array([
                    [0.95918367, 0.95918367, 0.95918367],
                    [0.97959184, 0.97959184, 0.97959184],
                    [1.        , 1.        , 1.        ]
                ]),
            index=['lag_3', 'lag_2', 'lag_1'],
            columns=["1", "2", "3"]
        ),
        None
    )

    pd.testing.assert_frame_equal(X_predict, expected[0])
    assert exog_predict is None


def test_create_predict_X_exog():
    """
    Test case for create_predict_X with exogenous variables
    """
    forecaster = ForecasterRnn(
        estimator=model_exog, levels=["1", "2", "3"], lags=10
    )
    forecaster.fit(series=series, exog=exog)

    warn_msg = re.escape(
        "The output matrix is in the transformed scale due to the "
        "inclusion of transformations in the Forecaster. "
        "As a result, any predictions generated using this matrix will also "
        "be in the transformed scale. Please refer to the documentation "
        "for more details: "
        "https://skforecast.org/latest/user_guides/training-and-prediction-matrices.html"
    )
    with pytest.warns(DataTransformationWarning, match = warn_msg):
        X_predict, exog_predict = forecaster.create_predict_X(steps=None, exog=exog_pred)

    expected = (
        pd.DataFrame(
            data=np.array([
                    [0.81632653, 0.81632653, 0.81632653],
                    [0.83673469, 0.83673469, 0.83673469],
                    [0.85714286, 0.85714286, 0.85714286],
                    [0.87755102, 0.87755102, 0.87755102],
                    [0.89795918, 0.89795918, 0.89795918],
                    [0.91836735, 0.91836735, 0.91836735],
                    [0.93877551, 0.93877551, 0.93877551],
                    [0.95918367, 0.95918367, 0.95918367],
                    [0.97959184, 0.97959184, 0.97959184],
                    [1.        , 1.        , 1.        ]
                ]),
            index=[f"lag_{i}" for i in range(10, 0, -1)],
            columns=["1", "2", "3"]
        ),
        pd.DataFrame(
            data=np.array([
                    [1.02040816, 1.02040816],
                    [1.04081633, 1.04081633],
                    [1.06122449, 1.06122449],
                    [1.08163265, 1.08163265],
                    [1.10204082, 1.10204082],
                    [1.12244898, 1.12244898],
                    [1.14285714, 1.14285714],
                    [1.16326531, 1.16326531]
                ]),
            index=[f"step_{i}" for i in range(1, 9)],
            columns=["exog1", "exog2"]
        )
    )

    pd.testing.assert_frame_equal(X_predict, expected[0])
    pd.testing.assert_frame_equal(exog_predict, expected[1])


def test_create_predict_X_specific_levels_with_exog():
    """
    Test case for create_predict_X with specific levels
    """
    forecaster = ForecasterRnn(
        estimator=model_exog, levels=["1", "2", "3"], lags=10
    )
    forecaster.fit(series=series, exog=exog)

    warn_msg = re.escape(
        "The output matrix is in the transformed scale due to the "
        "inclusion of transformations in the Forecaster. "
        "As a result, any predictions generated using this matrix will also "
        "be in the transformed scale. Please refer to the documentation "
        "for more details: "
        "https://skforecast.org/latest/user_guides/training-and-prediction-matrices.html"
    )
    with pytest.warns(DataTransformationWarning, match = warn_msg):
        X_predict, exog_predict = forecaster.create_predict_X(
            steps=5, exog=exog_pred, levels=["1", "2"]
        )

    expected = (
        pd.DataFrame(
            data=np.array([
                    [0.81632653, 0.81632653, 0.81632653],
                    [0.83673469, 0.83673469, 0.83673469],
                    [0.85714286, 0.85714286, 0.85714286],
                    [0.87755102, 0.87755102, 0.87755102],
                    [0.89795918, 0.89795918, 0.89795918],
                    [0.91836735, 0.91836735, 0.91836735],
                    [0.93877551, 0.93877551, 0.93877551],
                    [0.95918367, 0.95918367, 0.95918367],
                    [0.97959184, 0.97959184, 0.97959184],
                    [1.        , 1.        , 1.        ]
                ]),
            index=[f"lag_{i}" for i in range(10, 0, -1)],
            columns=["1", "2", "3"]
        ),
        pd.DataFrame(
            data=np.array([
                    [1.02040816, 1.02040816],
                    [1.04081633, 1.04081633],
                    [1.06122449, 1.06122449],
                    [1.08163265, 1.08163265],
                    [1.10204082, 1.10204082],
                    [1.12244898, 1.12244898],
                    [1.14285714, 1.14285714],
                    [1.16326531, 1.16326531]
                ]),
            index=[f"step_{i}" for i in range(1, 9)],
            columns=["exog1", "exog2"]
        )
    )

    pd.testing.assert_frame_equal(X_predict, expected[0])
    pd.testing.assert_frame_equal(exog_predict, expected[1])



# Unit test fit
# ==============================================================================
import pytest
import numpy as np
import pandas as pd
from skforecast.drift_detection import RangeDriftDetector


def test_fit_series_and_exog_single_series():
    """
    Test fit with series and exog as single Series.
    """
    series = pd.Series([1, 2, 3, 4, 5], name='y')
    exog = pd.Series([10, 20, 30, 40, 50], name='exog')
    detector = RangeDriftDetector()
    detector.fit(y=series, exog=exog)
    detector.__repr__()
    detector._repr_html_()

    assert detector.is_fitted is True
    assert detector.series_names_in_ == ['y']
    assert detector.series_values_range_ == {'y': (1.0, 5.0)}
    assert detector.exog_values_range_ == {'exog': (10.0, 50.0)}
    assert detector.exog_names_in_ == ['exog']


def test_fit_series_and_exog_df():
    """
    Test fit with series and exog as DataFrames.
    """
    series = pd.DataFrame({
        'series_1': [1, 2, 3, 4, 5],
        'series_2': [10, 20, 30, 40, 50]
    })
    exog = pd.DataFrame({
        'exog_1': [10, 20, 30, 40, 50],
        'exog_2': ['A', 'B', 'C', 'D', 'E']
    })
    detector = RangeDriftDetector()
    detector.fit(series=series, exog=exog)

    assert detector.is_fitted is True
    assert detector.series_names_in_ == ['series_1', 'series_2']
    assert detector.series_values_range_ == {
        'series_1': (1.0, 5.0),
        'series_2': (10.0, 50.0)
    }
    assert detector.exog_names_in_ == ['exog_1', 'exog_2']
    assert detector.exog_values_range_ == {
        'exog_1': (10.0, 50.0),
        'exog_2': set(['A', 'B', 'C', 'D', 'E'])
    }


def test_fit_series_and_exog_df_multindex():
    """
    Test fit with series and exog as DataFrames with MultiIndex.
    """
    index = pd.MultiIndex.from_product(
        [["series_1", "series_2", "series_3"], range(3)], names=["series", "time"]
    )
    series = pd.DataFrame({
        "value": [1, 2, 3, 10, 20, 30, 100, np.nan, 300]
    }, index=index)
    exog = pd.DataFrame(
        {
            "exog_1": [10, 20, 30, 100, 200, 300, 1000, np.nan, 3000],
            "exog_2": ["A", "B", "C", "D", "E", "A", "B", np.nan, "D"],
        },
        index=index,
    )

    detector = RangeDriftDetector()
    detector.fit(series=series, exog=exog)

    assert detector.is_fitted is True
    assert detector.series_names_in_ == ["series_1", "series_2", "series_3"]
    assert detector.series_values_range_ == {
        "series_1": (1.0, 3.0),
        "series_2": (10.0, 30.0),
        "series_3": (100.0, 300.0),
    }
    assert detector.exog_names_in_ == ["exog_1", "exog_2"]
    assert detector.exog_values_range_ == {
        "series_1": {
            "exog_1": (10, 30),
            "exog_2": {"A", "B", "C"}
        },
        "series_2": {
            "exog_1": (100, 300),
            "exog_2": {"A", "D", "E"}
        },
        "series_3": {
            "exog_1": (1000, 3000),
            "exog_2": {"B", "D"},
        },
    }

def test_fit_raise_error_series_and_y_provided():
    """
    Test fit raises TypeError when both series and y are provided.
    """
    series = pd.Series([1, 2, 3], name='y')
    detector = RangeDriftDetector()
    msg = (
        "Cannot specify both `series` and `y`. Please provide only one of them."
    )
    with pytest.raises(ValueError, match=msg):
        detector.fit(series=series, y=series)


def test_fit_raise_error_when_no_series_nor_y_provided():
    """
    Test fit raises TypeError when neither series nor y are provided.
    """
    detector = RangeDriftDetector()
    msg = (
        "One of `series` or `y` must be provided."
    )
    with pytest.raises(ValueError, match=msg):
        detector.fit(series=None, y=None)


def test_fit_invalid_series_type():
    """
    Test fit raises TypeError for invalid series type.
    """
    detector = RangeDriftDetector()
    msg = "Input must be a pandas Series, DataFrame or dict"
    with pytest.raises(TypeError, match=msg):
        detector.fit(series="invalid")


def test_fit_invalid_exog_type():
    """
    Test fit raises TypeError for invalid exog type.
    """
    series = pd.Series([1, 2, 3], name='y')
    detector = RangeDriftDetector()
    msg = "Exogenous variables must be a pandas Series, DataFrame or dict"
    with pytest.raises(TypeError, match=msg):
        detector.fit(series=series, exog="invalid")



# Unit test RollingFeatures
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from skforecast.preprocessing import RollingFeatures

# Fixtures
from .fixtures_preprocessing import X


def test_RollingFeatures_validate_params():
    """
    Test RollingFeatures _validate_params method.
    """

    params = {
        0: {'stats': 5, 'window_sizes': 5},
        1: {'stats': 'not_valid_stat', 'window_sizes': 5},
        2: {'stats': 'mean', 'window_sizes': 'not_int_list'},
        3: {'stats': 'mean', 'window_sizes': [5, 6]},
        4: {'stats': ['mean', 'median'], 'window_sizes': [6]},
        5: {'stats': ['mean', 'median', 'mean'], 'window_sizes': [6, 5, 6]},
        6: {'stats': ['mean'], 'window_sizes': [5], 'min_periods': 'not_int_list'},
        7: {'stats': ['mean'], 'window_sizes': 5, 'min_periods': [5, 4]},
        8: {'stats': ['mean', 'median'], 'window_sizes': 6, 'min_periods': [5]},
        9: {'stats': ['mean', 'median'], 'window_sizes': [5, 3], 'min_periods': [5, 4]},
        10: {'stats': ['mean', 'median'], 'window_sizes': [5, 6], 
             'min_periods': None, 'features_names': 'not_list'},
        11: {'stats': ['mean'], 'window_sizes': 5, 
             'min_periods': 5, 'features_names': ['mean_5', 'median_6']},
        12: {'stats': ['mean', 'median'], 'window_sizes': [5, 6], 
             'min_periods': 4, 'features_names': ['mean_5']},
        13: {'stats': ['mean', 'median'], 'window_sizes': [5, 6], 
             'min_periods': 5, 'features_names': ['mean_5', 'median_6'], 'fillna': {}},
        14: {'stats': ['mean', 'median'], 'window_sizes': [5, 6], 
             'min_periods': [5, 5], 'features_names': None, 'fillna': 'not_valid_fillna'},
        15: {'stats': ['mean', 'median'], 'window_sizes': [5, 6], 'min_periods': [5, 5], 
             'features_names': None, 'fillna': 'ffill', 'kwargs_stats': 'not_valid_kwargs_stats'},
        16: {'stats': ['mean', 'median'], 'window_sizes': [5, 6], 'min_periods': [5, 5], 
             'features_names': None, 'fillna': 'ffill', 'kwargs_stats': {'median': {'alpha': 0.3}}},
    }
    
    # stats
    err_msg = re.escape(
        f"`stats` must be a string or a list of strings. Got {type(params[0]['stats'])}."
    ) 
    with pytest.raises(TypeError, match = err_msg):
        RollingFeatures(**params[0])
    err_msg = re.escape(
        "Statistic 'not_valid_stat' is not allowed. Allowed stats are: ['mean', "
        "'std', 'min', 'max', 'sum', 'median', 'ratio_min_max', 'coef_variation', 'ewm']."
    ) 
    with pytest.raises(ValueError, match = err_msg):
        RollingFeatures(**params[1])

    # window_sizes
    err_msg = re.escape(
        f"`window_sizes` must be an int or a list of ints. Got {type(params[2]['window_sizes'])}."
    ) 
    with pytest.raises(TypeError, match = err_msg):
        RollingFeatures(**params[2])
    err_msg = re.escape(
        "Length of `window_sizes` list (2) must match length of `stats` list (1)."
    ) 
    with pytest.raises(ValueError, match = err_msg):
        RollingFeatures(**params[3])
    err_msg = re.escape(
        "Length of `window_sizes` list (1) must match length of `stats` list (2)."
    ) 
    with pytest.raises(ValueError, match = err_msg):
        RollingFeatures(**params[4])
    
    # Check duplicates (stats, window_sizes)
    err_msg = re.escape(
        "Duplicate (stat, window_size) pairs are not allowed.\n"
        "    `stats`        : ['mean', 'median', 'mean']\n"
        "    `window_sizes` : [6, 5, 6]"
    )
    with pytest.raises(ValueError, match = err_msg):
        RollingFeatures(**params[5])

    # min_periods
    err_msg = re.escape(
        f"`min_periods` must be an int, list of ints, or None. Got {type(params[6]['min_periods'])}."
    )
    with pytest.raises(TypeError, match = err_msg):
        RollingFeatures(**params[6])
    err_msg = re.escape(
        "Length of `min_periods` list (2) must match length of `stats` list (1)."
    ) 
    with pytest.raises(ValueError, match = err_msg):
        RollingFeatures(**params[7])
    err_msg = re.escape(
        "Length of `min_periods` list (1) must match length of `stats` list (2)."
    ) 
    with pytest.raises(ValueError, match = err_msg):
        RollingFeatures(**params[8])
    err_msg = re.escape(
        "Each `min_period` must be less than or equal to its corresponding `window_size`."
    ) 
    with pytest.raises(ValueError, match = err_msg):
        RollingFeatures(**params[9])

    # features_names
    err_msg = re.escape(
        f"`features_names` must be a list of strings or None. Got {type(params[10]['features_names'])}."
    )
    with pytest.raises(TypeError, match = err_msg):
        RollingFeatures(**params[10])
    err_msg = re.escape(
        "Length of `features_names` list (2) must match length of `stats` list (1)."
    ) 
    with pytest.raises(ValueError, match = err_msg):
        RollingFeatures(**params[11])
    err_msg = re.escape(
        "Length of `features_names` list (1) must match length of `stats` list (2)."
    ) 
    with pytest.raises(ValueError, match = err_msg):
        RollingFeatures(**params[12])

    # fillna
    err_msg = re.escape(
        f"`fillna` must be a float, string, or None. Got {type(params[13]['fillna'])}."
    )
    with pytest.raises(TypeError, match = err_msg):
        RollingFeatures(**params[13])
    err_msg = re.escape(
        "'not_valid_fillna' is not allowed. Allowed `fillna` "
        "values are: ['mean', 'median', 'ffill', 'bfill'] or a float value."
    ) 
    with pytest.raises(ValueError, match = err_msg):
        RollingFeatures(**params[14])

    # kwargs_stats
    err_msg = re.escape(
        f"`kwargs_stats` must be a dictionary or None. Got {type(params[15]['kwargs_stats'])}."
    )
    with pytest.raises(TypeError, match = err_msg):
        RollingFeatures(**params[15])
    allowed_kwargs_stats = ['ewm']
    err_msg = re.escape(
        f"Invalid statistic 'median' found in `kwargs_stats`. "
        f"Allowed statistics with additional arguments are: "
        f"{allowed_kwargs_stats}. Please ensure all keys in "
        f"`kwargs_stats` are among the allowed statistics."
    ) 
    with pytest.raises(ValueError, match = err_msg):
        RollingFeatures(**params[16])


@pytest.mark.parametrize(
    "params", 
    [{'stats': ['mean', 'ewm'], 'window_sizes': 5, 'min_periods': None, 
      'features_names': None, 'fillna': 'ffill', 
      'kwargs_stats': {'ewm': {'alpha': 0.3}}},
     {'stats': ['mean', 'ewm'], 'window_sizes': [5, 5], 'min_periods': 5, 
      'features_names': ['roll_mean_5', 'roll_ewm_5_alpha_0.3'], 
      'fillna': 'ffill', 'kwargs_stats': {'ewm': {'alpha': 0.3}}}], 
    ids = lambda params: f'params: {params}')
def test_RollingFeatures_init_store_parameters(params):
    """
    Test RollingFeatures initialization and stored parameters.
    """

    rolling = RollingFeatures(**params)

    assert rolling.stats == ['mean', 'ewm']
    assert rolling.n_stats == 2
    assert rolling.window_sizes == [5, 5]
    assert rolling.max_window_size == 5
    assert rolling.min_periods == [5, 5]
    assert rolling.features_names == ['roll_mean_5', 'roll_ewm_5_alpha_0.3']
    assert rolling.fillna == 'ffill'
    assert rolling.kwargs_stats == {'ewm': {'alpha': 0.3}}

    unique_rolling_windows = {
        '5_5': {'params': {'window': 5, 'min_periods': 5, 'center': False, 'closed': 'left'}, 
                'stats_idx': [0, 1],
                'stats_names': ['roll_mean_5', 'roll_ewm_5_alpha_0.3'],
                'rolling_obj': None}
    }

    assert rolling.unique_rolling_windows == unique_rolling_windows


def test_RollingFeatures_init_store_parameters_multiple_stats():
    """
    Test RollingFeatures initialization and stored parameters 
    when multiple stats are passed.
    """

    rolling = RollingFeatures(stats=['mean', 'median', 'sum'], window_sizes=[5, 5, 6])

    assert rolling.stats == ['mean', 'median', 'sum']
    assert rolling.n_stats == 3
    assert rolling.window_sizes == [5, 5, 6]
    assert rolling.max_window_size == 6
    assert rolling.min_periods == [5, 5, 6]
    assert rolling.features_names == ['roll_mean_5', 'roll_median_5', 'roll_sum_6']
    assert rolling.fillna is None

    unique_rolling_windows = {
        '5_5': {'params': {'window': 5, 'min_periods': 5, 'center': False, 'closed': 'left'}, 
                'stats_idx': [0, 1],
                'stats_names': ['roll_mean_5', 'roll_median_5'],
                'rolling_obj': None},
        '6_6': {'params': {'window': 6, 'min_periods': 6, 'center': False, 'closed': 'left'}, 
                'stats_idx': [2],
                'stats_names': ['roll_sum_6'],
                'rolling_obj': None}
    }

    assert rolling.unique_rolling_windows == unique_rolling_windows


def test_RollingFeatures_ValueError_apply_stat_when_stat_not_implemented():
    """
    Test RollingFeatures ValueError _apply_stat_pandas and _apply_stat_numpy 
    when applying a statistic not implemented.
    """
    
    rolling = RollingFeatures(stats='mean', window_sizes=10)
    X_window = X.iloc[-10:]
    rolling_obj = X_window.rolling(**rolling.unique_rolling_windows['10_10']['params'])

    err_msg = re.escape("Statistic 'not_valid' is not implemented.") 
    with pytest.raises(ValueError, match = err_msg):
        rolling._apply_stat_pandas(rolling_obj, 'not_valid')
    with pytest.raises(ValueError, match = err_msg):
        rolling._apply_stat_numpy_jit(X_window.to_numpy(), 'not_valid')


def test_RollingFeatures_apply_stat_pandas_numpy():
    """
    Test RollingFeatures _apply_stat_pandas and _apply_stat_numpy methods.
    """

    stats = ['mean', 'std', 'min', 'max', 'sum', 'median', 
             'ratio_min_max', 'coef_variation', 'ewm']
    
    rolling = RollingFeatures(stats=stats, window_sizes=10)
    X_window_pandas = X.iloc[-11:]
    X_window_numpy = X.to_numpy()[-11:-1]

    for stat in stats:

        rolling_obj = X_window_pandas.rolling(**rolling.unique_rolling_windows['10_10']['params'])
        stat_pandas = rolling._apply_stat_pandas(rolling_obj, stat).iat[-1]
        stat_numpy = rolling._apply_stat_numpy_jit(X_window_numpy, stat)

        np.testing.assert_almost_equal(stat_pandas, stat_numpy, decimal=7)


def test_RollingFeatures_transform_batch():
    """
    Test RollingFeatures transform_batch method.
    """
    X_datetime = X.copy()
    X_datetime.index = pd.date_range(start='1990-01-01', periods=len(X), freq='D')

    stats = ['mean', 'std', 'min', 'max', 'sum', 'median', 
             'ratio_min_max', 'coef_variation', 'ewm']
    rolling = RollingFeatures(stats=stats, window_sizes=4)
    rolling_features = rolling.transform_batch(X_datetime).head(10)

    expected = pd.DataFrame(
        data = {
            "roll_mean_4": [
                0.440193685,
                0.44594363000000004,
                0.48018541249999996,
                0.6686636,
                0.7020423425000001,
                0.642408075,
                0.63466084,
                0.475264295,
                0.4863192875,
                0.4757293725,
            ],
            "roll_std_4": [
                0.2215646508537583,
                0.2305486112172858,
                0.20796803421252608,
                0.24087135366452725,
                0.22810163043780196,
                0.25196045881621093,
                0.2612561344738438,
                0.15089728042696288,
                0.1715716346143639,
                0.17331344323200243,
            ],
            "roll_min_4": [
                0.22685145,
                0.22685145,
                0.22685145,
                0.42310646,
                0.42310646,
                0.42310646,
                0.39211752,
                0.34317802,
                0.34317802,
                0.34317802,
            ],
            "roll_max_4": [
                0.69646919,
                0.71946897,
                0.71946897,
                0.9807642,
                0.9807642,
                0.9807642,
                0.9807642,
                0.68482974,
                0.72904971,
                0.72904971,
            ],
            "roll_sum_4": [
                1.76077474,
                1.7837745200000001,
                1.9207416499999999,
                2.6746544,
                2.8081693700000003,
                2.5696323,
                2.53864336,
                1.90105718,
                1.94527715,
                1.90291749,
            ],
            "roll_median_4": [
                0.41872705,
                0.41872705,
                0.48721061499999996,
                0.63539187,
                0.702149355,
                0.58288082,
                0.58288082,
                0.43652471000000004,
                0.43652471000000004,
                0.41534488,
            ],
            "roll_ratio_min_max_4": [
                0.3257164182668296,
                0.3153040081770309,
                0.3153040081770309,
                0.43140487795129556,
                0.43140487795129556,
                0.43140487795129556,
                0.3998081496041556,
                0.5011143645718423,
                0.4707196440692638,
                0.4707196440692638,
            ],
            "roll_coef_variation_4": [
                0.5033344602700475,
                0.5169904797547747,
                0.4330994420046571,
                0.3602280035350021,
                0.3249114998185482,
                0.3922124715138597,
                0.4116468482187176,
                0.31750182375253516,
                0.35279627813314707,
                0.3643109995946328,
            ],
            "roll_ewm_4_alpha_0.3": [
                0.43000710180418467,
                0.5190257835333596,
                0.5032329347216739,
                0.7179547901342281,
                0.7206729972799052,
                0.6261400283379392,
                0.5529958717726017,
                0.42961449499407806,
                0.5236366168574812,
                0.4941020829688117,
            ],
        },
        index = pd.date_range(start='1990-01-05', periods=10, freq='D')
    )

    pd.testing.assert_frame_equal(rolling_features, expected)


def test_RollingFeatures_transform_batch_different_rolling_and_fillna():
    """
    Test RollingFeatures transform_batch method with different rolling windows 
    and fillna.
    """
    X_datetime = X.copy()
    X_datetime.index = pd.date_range(start='1990-01-01', periods=len(X), freq='D')
    X_datetime.iloc[9] = np.nan

    stats = ['mean', 'std', 'mean', 'max']
    window_sizes = [4, 5, 6, 4]
    min_periods = [3, 5, 6, 3]
    features_names = ['my_mean', 'my_std', 'my_mean_2', 'my_max']
    
    rolling = RollingFeatures(
        stats=stats, window_sizes=window_sizes, min_periods=min_periods,
        features_names=features_names, fillna='bfill'
    )
    rolling_features = rolling.transform_batch(X_datetime).head(15)
    
    expected = pd.DataFrame(
        data = np.array([
                   [0.48018541, 0.19992199, 0.4838917 , 0.71946897],
                   [0.6686636 , 0.28732186, 0.5312742 , 0.9807642 ],
                   [0.70204234, 0.20872596, 0.5977226 , 0.9807642 ],
                   [0.64240807, 0.22090887, 0.64006934, 0.9807642 ],
                   [0.71550861, 0.23906853, 0.45108626, 0.9807642 ],
                   [0.50297989, 0.23906853, 0.45108626, 0.68482974],
                   [0.51771988, 0.23906853, 0.45108626, 0.72904971],
                   [0.50359999, 0.23906853, 0.45108626, 0.72904971],
                   [0.39261947, 0.23906853, 0.45108626, 0.72904971],
                   [0.40633603, 0.23906853, 0.45108626, 0.72904971],
                   [0.40857245, 0.27992064, 0.45108626, 0.73799541],
                   [0.34455232, 0.2608389 , 0.42430521, 0.73799541],
                   [0.37349579, 0.26830576, 0.33203888, 0.73799541],
                   [0.40687257, 0.23934903, 0.3475354 , 0.73799541],
                   [0.35533061, 0.24575419, 0.42622702, 0.53182759]]),
        columns = ['my_mean', 'my_std', 'my_mean_2', 'my_max'],
        index = pd.date_range(start='1990-01-07', periods=15, freq='D')
    )

    pd.testing.assert_frame_equal(rolling_features, expected)


@pytest.mark.parametrize(
    "fillna", 
    ['mean', 'median', 'ffill', 'bfill', None, 5., 0], 
    ids = lambda fillna: f'fillna: {fillna}')
def test_RollingFeatures_transform_batch_fillna_all_methods(fillna):
    """
    Test RollingFeatures transform_batch method with all fillna methods.
    """
    X_datetime = X.head(10).copy()
    X_datetime.index = pd.date_range(start='1990-01-01', periods=len(X_datetime), freq='D')
    X_datetime.iloc[5] = np.nan

    base_array = np.array([0.40315332, 0.35476852, 0.49921173, 
                           np.nan, np.nan, np.nan, 0.715509])
    expected_dict = {
        'mean': np.array([0.49316055, 0.49316055, 0.49316055]),
        'median': np.array([0.45118253, 0.45118253, 0.45118253]),
        'ffill': np.array([0.49921173, 0.49921173, 0.49921173]),
        'bfill': np.array([0.71550861, 0.71550861, 0.71550861]),
        5.: np.array([5., 5., 5.]),
        0: np.array([0, 0, 0]),
    } 

    rolling = RollingFeatures(stats=['mean'], window_sizes=3, fillna=fillna)
    rolling_features = rolling.transform_batch(X_datetime)

    expected_array = base_array
    if fillna is not None:
        expected_array[-4:-1] = expected_dict[fillna]
    expected = pd.DataFrame(
        data=expected_array, columns=['roll_mean_3'], 
        index=pd.date_range(start='1990-01-04', periods=7, freq='D')
    )

    pd.testing.assert_frame_equal(rolling_features, expected)


def test_RollingFeatures_transform():
    """
    Test RollingFeatures transform method.
    """
    stats = [
        "mean",
        "std",
        "min",
        "max",
        "sum",
        "median",
        "ratio_min_max",
        "coef_variation",
        "ewm",
    ]
    rolling = RollingFeatures(stats=stats, window_sizes=4)
    rolling_features = rolling.transform(X.to_numpy(copy=True))

    expected = np.array(
        [
            0.65024343,
            0.23013666,
            0.48303426,
            0.98555979,
            2.6009737,
            0.56618983,
            0.49011157,
            0.35392385,
            0.64158672,
        ]
    )

    np.testing.assert_array_almost_equal(rolling_features, expected)


def test_RollingFeatures_transform_2d():
    """
    Test RollingFeatures transform method with 2 dimensions.
    """

    X_2d = X.to_numpy(copy=True)
    X_2d = np.tile(X_2d, (2, 1)).T

    stats = [
        "mean",
        "std",
        "min",
        "max",
        "sum",
        "median",
        "ratio_min_max",
        "coef_variation",
        "ewm",
    ]
    rolling = RollingFeatures(stats=stats, window_sizes=4)
    rolling_features = rolling.transform(X_2d)

    expected = np.array(
        [
            0.65024343,
            0.23013666,
            0.48303426,
            0.98555979,
            2.6009737,
            0.56618983,
            0.49011157,
            0.35392385,
            0.64158672,
        ]
    )
    expected = np.array([expected, expected])

    np.testing.assert_array_almost_equal(rolling_features, expected)


def test_RollingFeatures_transform_with_nans():
    """
    Test RollingFeatures transform method with nans.
    """
    
    X_nans = X.to_numpy(copy=True)
    X_nans[-7] = np.nan

    stats = [
        "mean",
        "std",
        "min",
        "max",
        "sum",
        "median",
        "ratio_min_max",
        "coef_variation",
        "ewm",
    ]
    window_sizes = [10, 10, 15, 4, 15, 4, 4, 4, 4]
    rolling = RollingFeatures(stats=stats, window_sizes=window_sizes)
    rolling_features = rolling.transform(X_nans)

    expected = np.array(
        [
            0.53051056,
            0.28145959,
            0.11561840,
            0.98555979,
            7.85259345,
            0.56618983,
            0.49011157,
            0.35392385,
            0.64158672,
        ]
    )

    np.testing.assert_array_almost_equal(rolling_features, expected)


def test_RollingFeatures_transform_with_nans_2d():
    """
    Test RollingFeatures transform method with nans and 2 dimensions.
    """
    
    X_2d_nans = X.to_numpy(copy=True)
    X_2d_nans = np.tile(X_2d_nans, (2, 1)).T
    X_2d_nans[-7, 0] = np.nan
    X_2d_nans[-5, 1] = np.nan

    stats = ['mean', 'std', 'min', 'max', 'sum', 'median', 
             'ratio_min_max', 'coef_variation']
    window_sizes = [10, 10, 15, 4, 15, 4, 4, 4]
    rolling = RollingFeatures(stats=stats, window_sizes=window_sizes)
    rolling_features = rolling.transform(X_2d_nans)
    
    expected_0 = np.array([0.53051056, 0.28145959, 0.11561840, 0.98555979, 
                           7.85259345, 0.56618983, 0.49011157, 0.35392385])
    expected_1 = np.array([0.548774, 0.26592, 0.115618, 0.98556, 
                           8.016964, 0.56618983, 0.49011157, 0.35392385])
    expected = np.array([expected_0, expected_1])

    np.testing.assert_array_almost_equal(rolling_features, expected)


def test_equivalence_results_RollingFeatures_and_custom_class():
    """
    Test equivalence of results between RollingFeatures and custom class that
    calculates rolling mean and std.
    """
        
    y = pd.Series([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    x = np.array([1, 20, 3])

    class RollingMeanStd():
        """
        Custom class to create rolling skewness features.
        """

        def __init__(self, window_sizes, features_names=['roll_mean_3', 'roll_std_3']):
            
            if not isinstance(window_sizes, list):
                window_sizes = [window_sizes]
            self.window_sizes = window_sizes
            self.features_names = features_names

        def transform_batch(self, X: pd.Series) -> pd.DataFrame:
            
            rolling_obj = X.rolling(window=self.window_sizes[0], center=False, closed='left')
            rolling_mean = rolling_obj.mean()
            rolling_std = rolling_obj.std()
            rolling_skewness = pd.DataFrame({
                                    'roll_mean_3': rolling_mean,
                                    'roll_std_3': rolling_std
                                }).dropna()

            return rolling_skewness

        def transform(self, X: np.ndarray) -> np.ndarray:
            
            X = X[~np.isnan(X)]
            if len(X) > 0:
                rolling_mean = np.mean(X)
                rolling_std = np.std(X, ddof=1)
                results = np.array([rolling_mean, rolling_std])
            else:
                results = np.nan
            
            return results
        
    window_features_1 = RollingFeatures(
                stats        = ['mean', 'std'],
                window_sizes = [3, 3]
                )
    window_features_2 = RollingMeanStd(window_sizes=3, features_names=['roll_mean_3', 'roll_std_3'])

    pd.testing.assert_frame_equal(
        window_features_1.transform_batch(y),
        window_features_2.transform_batch(y)
    )

    np.testing.assert_array_almost_equal(
        window_features_1.transform(x),
        window_features_2.transform(x)
    )



# Unit test _binning_out_sample_residuals ForecasterRecursiveMultiSeries
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from sklearn.linear_model import LinearRegression
from sklearn.preprocessing import StandardScaler
from skforecast.exceptions import ResidualsUsageWarning
from ....recursive import ForecasterRecursiveMultiSeries
from ....preprocessing import reshape_series_wide_to_long

# Fixtures
series = pd.DataFrame(
    {'l1': np.arange(10), 'l2': np.arange(10)},
    index = pd.date_range(start='1-1-2018', periods=10, freq='D')
)
series = reshape_series_wide_to_long(series)

rng = np.random.default_rng(12345)
series_rnd = pd.DataFrame(
    {"l1": rng.normal(10, 3, 20), "l2": rng.normal(10, 3, 20)},
    index = pd.date_range(start='01-01-2000', periods=20, freq='D')
)
series_rnd = reshape_series_wide_to_long(series_rnd)


@pytest.mark.parametrize("encoding", 
                         ['ordinal', 'onehot', 'ordinal_category'], 
                         ids=lambda encoding: f'encoding: {encoding}')
def test_binning_out_sample_residuals_when_residuals_length_is_less_than_10000_and_no_append(encoding):
    """
    Test residuals stored when new residuals length is less than 10000 and append
    is False.
    """
    rng = np.random.default_rng(12345)
    series_rnd = pd.DataFrame({"l1": rng.normal(10, 3, 20), "l2": rng.normal(10, 3, 20)})
    series_rnd.index = pd.date_range(start='01-01-2000', periods=len(series_rnd), freq='D')
    series_rnd = reshape_series_wide_to_long(series_rnd)

    forecaster = ForecasterRecursiveMultiSeries(
        LinearRegression(), lags=3, encoding=encoding, binner_kwargs={"n_bins": 3}
    )
    forecaster.fit(series=series_rnd)

    y_true = {"l1": rng.normal(10, 3, 20), "l2": rng.normal(10, 3, 20)}
    y_pred = {"l1": rng.normal(10, 3, 20), "l2": rng.normal(10, 3, 20)}

    forecaster.out_sample_residuals_ = {}
    forecaster.out_sample_residuals_by_bin_ = {}
    results = forecaster._binning_out_sample_residuals(
        level="l1", y_true=y_true["l1"], y_pred=y_pred["l1"]
    )

    expected_out_sample = np.array(
        [1.60962553,  1.71263736,  9.19708889, -2.73386424,  1.51464604,
        -3.32699998,  3.5987987 ,  0.13608979, -2.06018676, -2.72988393,
        1.63221968, -0.57130461, -1.08106429, -1.45886519,  0.91789988,
        -2.19933831,  1.97853178, -2.56217282,  3.69683321, -4.86497532]
    )
    expected_out_sample_by_bin = {
        0: np.array([
            1.60962553, 1.71263736, 9.19708889, 1.51464604, 3.5987987,
            1.63221968, -2.19933831, 1.97853178, -2.56217282, 3.69683321,
        ]),
        1: np.array([
            -3.32699998,
        ]),
        2: np.array([
            -2.73386424, 0.13608979, -2.06018676, -2.72988393, -0.57130461,
            -1.08106429, -1.45886519, 0.91789988, -4.86497532,
        ])
    }
    
    np.testing.assert_array_almost_equal(results[0], expected_out_sample)
    assert results[1].keys() == expected_out_sample_by_bin.keys()
    assert results[1].keys() == expected_out_sample_by_bin.keys()
    for k in results[1].keys():
        np.testing.assert_array_almost_equal(
            results[1][k], expected_out_sample_by_bin[k]
        )


def test_binning_out_sample_residuals_when_residuals_length_is_less_than_10000_and_no_append_encoding_None():
    """
    Test residuals stored when new residuals length is less than 10000 and append
    is False and encoding is None.
    """
    rng = np.random.default_rng(12345)
    series_rnd = pd.DataFrame({"l1": rng.normal(10, 3, 20), "l2": rng.normal(10, 3, 20)})
    series_rnd.index = pd.date_range(start='01-01-2000', periods=len(series_rnd), freq='D')
    series_rnd = reshape_series_wide_to_long(series_rnd)
    
    forecaster = ForecasterRecursiveMultiSeries(
        LinearRegression(), lags=3, encoding=None, binner_kwargs={"n_bins": 3}
    )
    forecaster.fit(series=series_rnd)

    y_true = {"_unknown_level": pd.Series(rng.normal(10, 3, 20))}
    y_pred = {"_unknown_level": pd.Series(rng.normal(10, 3, 20))}

    forecaster.out_sample_residuals_ = {}
    forecaster.out_sample_residuals_by_bin_ = {}
    results = forecaster._binning_out_sample_residuals(
        level="_unknown_level", y_true=y_true["_unknown_level"], y_pred=y_pred["_unknown_level"]
    )

    expected_out_sample = np.array(
        [ 1.15252117, -1.81458426,  5.05451064, -0.57203687,  4.72280084,
         -0.75286693, -1.65697481,  7.50482204, -1.7679251 ,  4.37782366,
          4.02311098,  4.1187269 ,  1.56932008,  0.0070832 ,  5.05247358,
         -1.84063415, -3.96576256, -3.85592894,  3.24784139, -4.9446084]
    )
    expected_out_sample_by_bin = {
        0: np.array([
            5.05451064,  4.72280084, -0.75286693,  7.50482204,  4.37782366,
            4.02311098,  4.1187269 ,  5.05247358, -1.84063415, -3.85592894,
            3.24784139
        ]),
        1: np.array([
            1.15252117, -0.57203687,  1.56932008
        ]),
        2: np.array([
            -1.81458426, -1.65697481, -1.7679251 ,  0.0070832 , -3.96576256,
            -4.9446084
        ])
    }
    
    np.testing.assert_array_almost_equal(results[0], expected_out_sample)
    assert results[1].keys() == expected_out_sample_by_bin.keys()
    assert results[1].keys() == expected_out_sample_by_bin.keys()
    for k in results[1].keys():
        np.testing.assert_array_almost_equal(
            results[1][k], expected_out_sample_by_bin[k]
        )


@pytest.mark.parametrize("encoding", ['ordinal', 'onehot', 'ordinal_category'], 
                         ids=lambda encoding: f'encoding: {encoding}')
def test_binning_out_sample_residuals_when_residuals_length_is_less_than_10000_and_append(encoding):
    """
    Test residuals stored when new residuals length is less than 10_000 and append is True.
    """
    forecaster = ForecasterRecursiveMultiSeries(
        LinearRegression(), lags=3, encoding=encoding, binner_kwargs={'n_bins': 3}
    )
    forecaster.fit(series=series)
    y_true = {'l1': np.array([1, 2, 3, 4, 5]), 'l2': np.array([2, 3, 4, 5, 6])}
    y_pred = {'l1': np.array([0, 1, 2, 3, 4]), 'l2': np.array([0, 1, 2, 3, 4])}
    
    forecaster.out_sample_residuals_ = {}
    forecaster.out_sample_residuals_by_bin_ = {}

    warn_msg = re.escape(
        f"The following bins of level 'l1' have no out of sample residuals: [1, 2]. "
        f"No predicted values fall in the interval {[forecaster.binner_intervals_['l1'][bin] for bin in [1, 2]]}. "
        f"Empty bins will be filled with a random sample of residuals."
    )
    with pytest.warns(ResidualsUsageWarning, match=warn_msg):
        residuals = forecaster._binning_out_sample_residuals(
            level="l1", y_true=y_true["l1"], y_pred=y_pred["l1"]
        )
    forecaster.out_sample_residuals_ = {"l1": residuals[0]}
    forecaster.out_sample_residuals_by_bin_ = {"l1": residuals[1]}

    y_true = {'l1': np.array([2, 3, 4, 5, 6]), 'l2': np.array([2, 3, 4, 5, 6])}
    y_pred = {'l1': np.array([0, 1, 2, 3, 4]), 'l2': np.array([0, 1, 2, 3, 4])}
    results = forecaster._binning_out_sample_residuals(
        level="l1", y_true=y_true["l1"], y_pred=y_pred["l1"], append=True
    )

    expected_out_sample = np.array([1, 1, 1, 1, 1, 2, 2, 2, 2, 2])
    expected_out_sample_by_bin = {
        0: np.array([1, 1, 1, 1, 1, 2, 2, 2, 2, 2]), 
        1: np.array([1, 1, 1, 1, 1]),
        2: np.array([1, 1, 1, 1, 1])
    }

    np.testing.assert_array_almost_equal(results[0], expected_out_sample)
    assert results[1].keys() == expected_out_sample_by_bin.keys()
    assert results[1].keys() == expected_out_sample_by_bin.keys()
    for k in results[1].keys():
        np.testing.assert_array_almost_equal(
            results[1][k], expected_out_sample_by_bin[k]
        )


@pytest.mark.parametrize("encoding", 
                         ['ordinal', 'onehot', 'ordinal_category'], 
                         ids=lambda encoding: f'encoding: {encoding}')
def test_binning_out_sample_residuals_when_residuals_length_is_greater_than_10000(encoding):
    """
    Test len residuals stored when its length is greater than 10000.
    """
    forecaster = ForecasterRecursiveMultiSeries(
        LinearRegression(), lags=3, encoding=encoding, binner_kwargs={'n_bins': 2}
    )
    forecaster.fit(series=series)
    y_true = {'l1': np.ones(20_000)}
    y_pred = {'l1': np.concatenate([np.ones(10_000) + 1, np.ones(10_000) + 2])}
    
    forecaster.out_sample_residuals_ = {}
    forecaster.out_sample_residuals_by_bin_ = {}
    results = forecaster._binning_out_sample_residuals(
        level="l1", y_true=y_true["l1"], y_pred=y_pred["l1"]
    )

    assert len(results[0]) == 10_000
    for v in results[1].values():
        assert len(v) == 5_000


def test_binning_out_sample_residuals_when_residuals_length_is_greater_than_10000_encoding_None():
    """
    Test len residuals stored when its length is greater than 10_000
    and encoding is None.
    """
    forecaster = ForecasterRecursiveMultiSeries(
        LinearRegression(), lags=3, encoding=None,  binner_kwargs={'n_bins': 2}
    )
    forecaster.fit(series=series)
    y_true = {'_unknown_level': np.ones(20_000)}
    y_pred = {'_unknown_level': np.concatenate([np.ones(10_000) + 1, np.ones(10_000) + 2])}
    
    forecaster.out_sample_residuals_ = {}
    forecaster.out_sample_residuals_by_bin_ = {}
    results = forecaster._binning_out_sample_residuals(
        level="_unknown_level", y_true=y_true["_unknown_level"], y_pred=y_pred["_unknown_level"]
    )

    assert len(results[0]) == 10_000
    for v in results[1].values():
        assert len(v) == 5_000


def test_binning_out_sample_residuals_when_residuals_length_is_greater_than_10000_and_append():
    """
    Test residuals stored when new residuals length is greater than 10000 and
    append is True.
    """
    forecaster = ForecasterRecursiveMultiSeries(
        LinearRegression(), lags=3, encoding='ordinal',  binner_kwargs={'n_bins': 2}
    )
    forecaster.fit(series=series)
    y_true = {'l1': np.ones(5_000)}
    y_pred = {'l1': np.concatenate([np.ones(2_500) + 1, np.ones(2_500) + 2])}

    forecaster.out_sample_residuals_ = {}
    forecaster.out_sample_residuals_by_bin_ = {}
    results = forecaster._binning_out_sample_residuals(
        level="l1", y_true=y_true["l1"], y_pred=y_pred["l1"]
    )
    
    y_true = {'l1': np.ones(20_000)}
    y_pred = {'l1': np.concatenate([np.ones(10_000) + 1, np.ones(10_000) + 2])}
    results = forecaster._binning_out_sample_residuals(
        level="l1", y_true=y_true["l1"], y_pred=y_pred["l1"], append=True
    )

    assert len(results[0]) == 10_000
    for v in results[1].values():
        assert len(v) == 5_000


@pytest.mark.parametrize("differentiation", 
                         [1, {'l1': 1, 'l2': 1, '_unknown_level': 1}], 
                         ids=lambda diff: f'differentiation: {diff}')
def test_forecaster_binning_out_sample_residuals_when_transformer_series_and_differentiation(differentiation):
    """
    Test _binning_out_sample_residuals when forecaster has transformer_series and differentiation.
    Stored should equivalent to residuals calculated manually if transformer_series and
    differentiation are applied to `y_true` and `y_pred` before calculating residuals.
    """
    series_train = {
        'l1': pd.Series(
            np.array([-1.42382504,  1.26372846, -0.87066174, -0.25917323, -0.07534331,
                      -0.74088465, -1.3677927 ,  0.6488928 ,  0.36105811, -1.95286306,
                       2.34740965,  0.96849691, -0.75938718,  0.90219827, -0.46695317,
                      -0.06068952,  0.78884434, -1.25666813,  0.57585751,  1.39897899]),
            index = pd.date_range(start='1-1-2018', periods=20, freq='D')
        ),
        'l2': pd.Series(
            np.array([1.32229806, -0.29969852,  0.90291934, -1.62158273, -0.15818926,
                      0.44948393, -1.34360107, -0.08168759,  1.72473993,  2.61815943,
                      0.77736134,  0.8286332 , -0.95898831, -1.20938829, -1.41229201,
                      0.54154683,  0.7519394 , -0.65876032, -1.22867499,  0.25755777]),
            index = pd.date_range(start='1-1-2018', periods=20, freq='D')
        )
    }
    y_true  = {
        'l1': np.array([ 0.31290292, -0.13081169,  1.26998312, -0.09296246, -0.06615089]),
        'l2': np.array([-1.10821447,  0.13595685,  1.34707776,  0.06114402,  0.0709146 ])
    }
    y_pred = {
        'l1': np.array([0.43365454, 0.27748366, 0.53025239, 0.53672097, 0.61835001]),
        'l2': np.array([-0.79501746,  0.30003095, -1.60270159,  0.26679883, -1.26162378])
    }
    
    forecaster = ForecasterRecursiveMultiSeries(
                     estimator          = LinearRegression(),
                     lags               = 5,
                     transformer_series = StandardScaler(),
                     differentiation    = differentiation
                 )
    forecaster.fit(series=series_train)
    forecaster.out_sample_residuals_ = {}
    forecaster.out_sample_residuals_by_bin_ = {}
    results = forecaster._binning_out_sample_residuals(
        level="l1", y_true = y_true['l1'], y_pred = y_pred['l1']
    )

    y_true['l1'] = forecaster.transformer_series_['l1'].transform(y_true['l1'].reshape(-1, 1)).flatten()
    y_pred['l1'] = forecaster.transformer_series_['l1'].transform(y_pred['l1'].reshape(-1, 1)).flatten()
    y_true['l1'] = forecaster.differentiator_['l1'].transform(y_true['l1'])[forecaster.differentiation_max:]
    y_pred['l1'] = forecaster.differentiator_['l1'].transform(y_pred['l1'])[forecaster.differentiation_max:]
    residuals_l1 = y_true['l1'] - y_pred['l1']

    np.testing.assert_array_almost_equal(results[0], residuals_l1)



# Unit test _create_lags ForecasterDirect
# ==============================================================================
import numpy as np
import pandas as pd
from sklearn.linear_model import LinearRegression
from skforecast.preprocessing import RollingFeatures
from skforecast.direct import ForecasterDirect

rolling = RollingFeatures(
    stats=['mean', 'median', 'sum'], window_sizes=[5, 5, 6]
)

  
def test_create_lags_when_lags_is_3_steps_1_and_y_is_numpy_arange_10():
    """
    Test matrix of lags created properly when lags is 3, steps is 1 and y is
    np.arange(10).
    """
    forecaster = ForecasterDirect(
        estimator=LinearRegression(), lags=3, steps=1
    )
    results = forecaster._create_lags(y=np.arange(10))
    expected = (
        np.array([[2., 1., 0.],
                  [3., 2., 1.],
                  [4., 3., 2.],
                  [5., 4., 3.],
                  [6., 5., 4.],
                  [7., 6., 5.],
                  [8., 7., 6.]]),
        np.array(
            [[3.], 
             [4.], 
             [5.], 
             [6.], 
             [7.], 
             [8.], 
             [9.]]
        )
    )

    np.testing.assert_array_almost_equal(results[0], expected[0])
    np.testing.assert_array_almost_equal(results[1], expected[1])


def test_create_lags_output_pandas():
    """
    Test matrix of lags is created properly when X_as_pandas=True.
    """
    forecaster = ForecasterDirect(
        estimator=LinearRegression(), lags=3, steps=1
    )
    results = forecaster._create_lags(
        y=np.arange(10, dtype=float), X_as_pandas=True, 
        train_index=pd.date_range('2020-01-03', periods=7, freq='D')
    )
    expected = (
        pd.DataFrame(
            data = np.array([
                       [2., 1., 0.],
                       [3., 2., 1.],
                       [4., 3., 2.],
                       [5., 4., 3.],
                       [6., 5., 4.],
                       [7., 6., 5.],
                       [8., 7., 6.]]
                   ),
            columns = ["lag_1", "lag_2", "lag_3"],
            index = pd.date_range('2020-01-03', periods=7, freq='D')
        ),
        np.array(
            [[3.], 
             [4.], 
             [5.], 
             [6.], 
             [7.], 
             [8.], 
             [9.]]
        )
    )

    pd.testing.assert_frame_equal(results[0], expected[0])
    np.testing.assert_array_almost_equal(results[1], expected[1])


def test_create_lags_when_lags_is_list_interspersed_lags_steps_1_and_y_is_numpy_arange_10():
    """
    Test matrix of lags created properly when lags is a list with interspersed 
    lags, steps is 1 and y is np.arange(10).
    """
    forecaster = ForecasterDirect(
        estimator=LinearRegression(), lags=[1, 5], steps=1
    )
    results = forecaster._create_lags(y=np.arange(10))
    expected = (
        np.array([[4., 0.],
                  [5., 1.],
                  [6., 2.],
                  [7., 3.],
                  [8., 4.]]),
        np.array(
            [[5.], 
             [6.], 
             [7.], 
             [8.], 
             [9.]]
        )
    )

    np.testing.assert_array_almost_equal(results[0], expected[0])
    np.testing.assert_array_almost_equal(results[1], expected[1])


def test_create_lags_when_lags_is_3_steps_2_and_y_is_numpy_arange_10():
    """
    Test matrix of lags created properly when lags is 3, steps is 2 and y is
    np.arange(10).
    """
    forecaster = ForecasterDirect(
        estimator=LinearRegression(), lags=3, steps=2
    )
    results = forecaster._create_lags(y=np.arange(10))
    expected = (
        np.array([[2., 1., 0.],
                  [3., 2., 1.],
                  [4., 3., 2.],
                  [5., 4., 3.],
                  [6., 5., 4.],
                  [7., 6., 5.]]),
        np.array([
            [3., 4.],
            [4., 5.],
            [5., 6.],
            [6., 7.],
            [7., 8.],
            [8., 9.]]
        )
    )

    np.testing.assert_array_almost_equal(results[0], expected[0])
    np.testing.assert_array_almost_equal(results[1], expected[1])


def test_create_lags_when_lags_is_3_steps_5_and_y_is_numpy_arange_10():
    """
    Test matrix of lags created properly when lags is 3, steps is 5 and y is
    np.arange(10).
    """
    forecaster = ForecasterDirect(
        estimator=LinearRegression(), lags=3, steps=5
    )
    results = forecaster._create_lags(y=np.arange(10))
    expected = (
        np.array([[2., 1., 0.],
                  [3., 2., 1.],
                  [4., 3., 2.]]),
        np.array([[3., 4., 5., 6., 7.],
                  [4., 5., 6., 7., 8.],
                  [5., 6., 7., 8., 9.]])
    )

    np.testing.assert_array_almost_equal(results[0], expected[0])
    np.testing.assert_array_almost_equal(results[1], expected[1])


def test_create_lags_output_lags_None():
    """
    Test matrix of lags when lags=None.
    """
    forecaster = ForecasterDirect(
        estimator=LinearRegression(), lags=None, steps=2, window_features=rolling
    )
    results = forecaster._create_lags(y=np.arange(10))
    expected = (
        None,
        np.array([[6., 7.],
                  [7., 8.],
                  [8., 9.]])
    )

    assert results[0] == expected[0]
    np.testing.assert_array_almost_equal(results[1], expected[1])


def test_create_lags_when_window_size_window_features_greater_than_max_lag():
    """
    Test matrix of lags created properly when lags is 3, steps is 2, y is
    np.arange(10) and window_size of window_features is greater than max lag.
    """
    forecaster = ForecasterDirect(
        estimator=LinearRegression(), lags=3, steps=2, window_features=rolling
    )
    results = forecaster._create_lags(y=np.arange(10))
    expected = (
        np.array([[5., 4., 3.],
                  [6., 5., 4.],
                  [7., 6., 5.]]),
        np.array([
            [6., 7.],
            [7., 8.],
            [8., 9.]]
        )
    )

    np.testing.assert_array_almost_equal(results[0], expected[0])
    np.testing.assert_array_almost_equal(results[1], expected[1])
